/* ============================================================================
 * File Name: DDS_sw_01a_dev
 *   DDS_sw v0.0 custom component
 *
 * Description:
 *   Implements software DDS to generate sine, ramp and triangle waves
 *   Output data types:
 *      int8, int16, int32, uint8, uint16, uin32, float
 *
 *
 * Credits:
 *
 * ============================================================================
 * PROVIDED AS-IS, NO WARRANTY OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * FREE TO SHARE, USE AND MODIFY UNDER TERMS: CREATIVE COMMONS - SHARE ALIKE
 * ============================================================================
*/


#include <`$INSTANCE_NAME`.h> // must specify API prefix in Symbol->Properties->Doc.APIPrefix
#include <`$INSTANCE_NAME`_Table.h>
#include <`$INSTANCE_NAME`_sine10.h>
#include <math.h>

//====================================
//        private definitions
//====================================
//#define `$INSTANCE_NAME`_DataPacketSize   `=$data_packet_size`  // size of sinle data packet (slice)  

#define _sine8      (0x01u)
#define _sine12     (0x02u)
#define _math       (0x03u)
    
#define _int32      (0x01u)
#define _uint32     (0x02u)


#define DDS_RESOLUTION (double) 4294967296.0    // 4294967296 = 2^32 
#define TUNE_WORD_MAX  (uint32) 2147483647u     //  2147483647 = 2^23 - 1 = DDS_RESOLUTION/2 - 1


//====================================
//        private variables
//====================================
//static double CLOCK_FREQ = `=$ClockFreq`;     // DDS sampling clock frequency
static double aSamplingFreq = `=$ClockFreq`;    // DDS sampling clock preset frequency
static double aFrequency    = `=$Frequency`;    // preset output frequency  
static double Tdiv;		                        // time period per bit increment
static int32  aAmplitude    = `=$Amplitude`;    // DDS wave amplitude 

static uint32 aDDS_acc;                         // DDS accumulator
static uint32 aDDS_inc;                         // DDS tune word inctrement 

static uint8  aEnabled = false;                 // DDS enabled status 

  
//====================================
//        private functions
//====================================



//==============================================================================
// Get next wave data sample.
// input nextval=false - no increment, nextval = true - next step
//==============================================================================

`$data_t` `$INSTANCE_NAME`_GetSample(uint8 nextval)
{     
    if (!aEnabled) return 0;
    
    //const uint16 half_pi = 256; // table len=1024
    //const uint16 half_pi = 64;

    `$data_t` val1=0;     // sin value
    `$data_t` val2=0;     // cos value
    uint32  phase=0;

    if (nextval!=0)
        aDDS_acc += aDDS_inc;   // increment DDS accumulator


    #if   (`$INSTANCE_NAME`_Algorithm == _sine8)

        //aAmplitude=5; //debug.
        const uint16 half_pi = 64;

        phase = aDDS_acc;
        //phase -= (aDDS_inc>>1); // shift by half-step
        phase >>= 24;   // 8-bit table
        //phase = (aDDS_acc >> 24);   // 8-bit table

       
        #if   (`$INSTANCE_NAME`_DataType == _int32)
            int32 offset = (aAmplitude & 0x01)? 0 : 255;
            val1 = ((offset + aAmplitude * ((`$INSTANCE_NAME`_sine8[(uint8) phase]         <<1)-255)) >> 9); // output = [-128..127]
            val2 = ((offset + aAmplitude * ((`$INSTANCE_NAME`_sine8[(uint8)(phase+half_pi)]<<1)-255)) >> 9);      
        #else
            //int32 offset = 128;
            //val1 = ((offset + aAmplitude * (`$INSTANCE_NAME`_sine8[(uint8) phase]         )) >> 8); // sin [0..255] 
            //val2 = ((offset + aAmplitude * (`$INSTANCE_NAME`_sine8[(uint8)(phase+half_pi)])) >> 8); // cos [0..255]        
            //val1 = (128 + aAmplitude * (`$INSTANCE_NAME`_sine8[(uint8)(phase)]         +0)) >>8; // sin [0..255] 
            //val2 = (255 + aAmplitude * (`$INSTANCE_NAME`_sine8[(uint8)(phase+half_pi)] +0)) >>8; // cos [0..255]        
               
            int32 offset = (aAmplitude & 0x01)? 0 : 255;
            int32 offset2= (aAmplitude+1)>>1;
            val1 = offset2 + ((offset + aAmplitude * ((`$INSTANCE_NAME`_sine8[(uint8) phase]         <<1)-255)) >> 9);  //output = [0..255]
            val2 = offset2 + ((offset + aAmplitude * ((`$INSTANCE_NAME`_sine8[(uint8)(phase+half_pi)]<<1)-255)) >> 9); 
        #endif // _sine8

        
    #elif (`$INSTANCE_NAME`_Algorithm == _sine12)
    
        //aAmplitude=5; //debug.
        const uint16 half_pi = 256; // table size 1024
        
        //phase = (aDDS_acc >> 22);   // table size 1024
        phase = aDDS_acc;
        phase -= (aDDS_inc>>1); // half-step only
        phase >>= 22;   // 11-bit table

        #if   (`$INSTANCE_NAME`_DataType == _int32)
            //int32 offset = (aAmplitude & 0x01)? 0 : 1024;
            //val1 = (int32) (offset + aAmplitude * mysine(phase)) >> 11; // [-128..127]
            //val2 = (int32) (offset + aAmplitude * mysine(phase+half_pi)) >> 11; // [-128..127]
            
            //int32 offset = (aAmplitude+1)>>1;
            //val1 = -0 + ((int32) (offset + aAmplitude * mysine(phase) * 265/256) >>11); //sin - OK [-128..127]
            ///val1 = -0 + ((int32) (0 +  mysine(phase)*265/256  ) >>3); //sin[-128..127]
            
            //half-step only->
            ///int32 offset = (aAmplitude & 0x01)? 0 : 1;
            int32 offset = !(aAmplitude & 0x01); // correction, using "sine_A1024_N1023_HS_signed"
            val1 = (int32) (( offset+ ((aAmplitude * mysine(phase) * 265/256)>>10)  )>>1); //sin, half-step only, OK [-128..127]

            // full step only->
            ///int32 offset = (aAmplitude & 0x01)? 0 : 1;
            //int32 offset = !(aAmplitude & 0x01); // correction, using "sine_A1024_N1023_HS_signed"
            //val1 = (int32) (( offset+ ((aAmplitude * mysine(phase) )>>10)  )>>1); //sin, half-step only, OK [-128..127]
            

            // table size 2048
            //uint16 phase = (aDDS_acc >> 21);   // for table size 2^11=2048
            //const uint16 half_pi = 512;
            //int32 offset = (aAmplitude & 0x01)? 0 : 1024;
            //int16 val1 = (int32) (offset + aAmplitude * mysine(phase)) >> 11; // [-128..127]
            //int16 val2 = (int32) (offset + aAmplitude * mysine(phase+half_pi)) >> 11; // [-128..127]
        #else
            int32 offset = 1024;
            val1 = (uint32) (offset + aAmplitude * (mysine(phase)+1024)) >> 11; // [0..255] // using signed mysine
            val2 = (uint32) (offset + aAmplitude * (mysine(phase+half_pi)+offset)) >> 11; // cosin [0..255] // using signed mysine
        #endif // _sine12
        
        
    #elif (`$INSTANCE_NAME`_Algorithm == _math)
    
        //aAmplitude=5; // debug
        phase = aDDS_acc;
        //phase -=( aDDS_inc>>1); // half-step only
        float32 phase_f = (float32) (2 * M_PI * phase / DDS_RESOLUTION);

        float32 fAmplitude = (float) aAmplitude; //full-step only->
        //float32 fAmplitude = (float) aAmplitude * (265.0/255.0); // half-step only
        
        #if   (`$INSTANCE_NAME`_DataType == _int32)
            //val1 = 0.5*aAmplitude * sin(phase_f);
            //val2 = 0.5*aAmplitude * cos(phase_f);
            //val1 = round(aAmplitude * sin(Phase));
            //val2 = round(0.5*aAmplitude * cos(Phase));

            int32 offset = (aAmplitude+1)>>1;
            //val1 = -offset + ((int32) (1 + aAmplitude + aAmplitude * sin(phase_f)) >>1); // sin
            //val2 = -offset + ((int32) (1 + aAmplitude + aAmplitude * cos(phase_f)) >>1); //cos
            val1 = -offset + ((int32) (1 + aAmplitude + fAmplitude * sin(phase_f)) >>1); //sin - OK [-128..127]
            //val1 = ((int32) (0.0 + fAmplitude * sin(phase_f)) /2.0); // sin
        #else        
            //val1 = (uint32) (1 + aAmplitude + aAmplitude * sin(phase_f)) >>1; // normal step
            //val2 = (uint32) (1 + aAmplitude + aAmplitude * cos(phase_f)) >>1;
            val1 = (uint32) (1 + aAmplitude + fAmplitude * sin(phase_f)) >>1; // half-step only OK [0..255]
            
            ///val1 = (val1>0xFF)? 0xFF : (val1>0x7FFFFFFF)? 0x00 : val1;

        #endif // _math    

    #endif
    

    `$INSTANCE_NAME`_sin = val1;
    `$INSTANCE_NAME`_cos = val2;
    `$INSTANCE_NAME`_phase = phase;
    //`$INSTANCE_NAME`_ref = (phase >> 31); // msb 
    //`$INSTANCE_NAME`_ref = (((phase >> 31) << 1) - 1) << 16; // uint 1:0 
    //`$INSTANCE_NAME`_ref = -(((phase >> 31) << 1) - 1) << 16; // int -1:1 x amplitude 
    `$INSTANCE_NAME`_ref = aAmplitude * (((phase >> 31) << 1) - 1); // int -1:1 x aAmplitude 
    
    return val1;
}

//==============================================================================
// Initialize component
// Set input clock frequency and default values for output frequency and phase 
//==============================================================================
void `$INSTANCE_NAME`_Init()
{   
    // make record of the input clock frequency into period   
    //Tdiv = (double) (DDS_RESOLUTION / CLOCK_FREQ);	
    Tdiv = (double) (DDS_RESOLUTION / aSamplingFreq);	

    
    aDDS_acc = 0;

    //`$INSTANCE_NAME`_SetFrequency( `$INSTANCE_NAME`_PresetFreq ); // todo: Frequency?
    `$INSTANCE_NAME`_SetFrequency( aFrequency ); // todo: Frequency?
    //aInitialized = true;
    //aEnabled = true;
}

//==============================================================================
// Enable component
// Have effect only in Software_only and Software_and_Hardware modes 
//==============================================================================

void `$INSTANCE_NAME`_Enable()
{   
    //todo: must be initialized first
    aEnabled = true;
}

//==============================================================================
// Start component
// Initialize and enable DDS component
//==============================================================================

void `$INSTANCE_NAME`_Start()
{   
    `$INSTANCE_NAME`_Init();    
    `$INSTANCE_NAME`_Enable();
}


//==============================================================================
// Stop component
//==============================================================================
void `$INSTANCE_NAME`_Stop()
{ 
    aEnabled = false;
}


//==============================================================================
// Set output frequency
//==============================================================================
uint8 `$INSTANCE_NAME`_SetFrequency( double freq )
{     
    uint32 _inc = (uint32) (freq * Tdiv + 0.5);             // calculate tune word trial value
    if ( (_inc < 1) || (_inc > TUNE_WORD_MAX) )  return 0;  // fail -> exit if outside of the valid range 
          
    aDDS_inc   = _inc;      // set tune word
    aFrequency = freq;      // backup value

    return 1; // success
}

//==============================================================================
// Set sampling frequency
//==============================================================================
uint8 `$INSTANCE_NAME`_SetClock( double freq )
{       
    // make record of the input clock frequency into period   
    double _Tdiv = (double) (DDS_RESOLUTION / freq);	    // trial
 
    uint32 _inc = (uint32) (aFrequency * _Tdiv + 0.5);      // calculate tune word trial value
    if ( (_inc < 1) || (_inc > TUNE_WORD_MAX) )  return 0;  // fail -> exit if outside of the valid range 
    
    Tdiv = _Tdiv;           // set divider
    aDDS_inc = _inc;        // set tune word
    aSamplingFreq = freq;   // backup value
    
    return 1; // success
}

//==============================================================================
// Set output amplitude
//==============================================================================
void `$INSTANCE_NAME`_SetAmplitude( int32 value )
{     
    aAmplitude = value; 
}

//==============================================================================
// Get output amplitude
//==============================================================================
int32 `$INSTANCE_NAME`_GetAmplitude() { return aAmplitude; }

//==============================================================================
// Get current sampling clock frequency
//==============================================================================
double `$INSTANCE_NAME`_GetClock() { return aSamplingFreq; }



//==============================================================================
// Reset DDS accumulator
//==============================================================================
void `$INSTANCE_NAME`_Reset()
{  
    aDDS_acc = 0;
}


//==============================================================================
//     Helper functions->
//==============================================================================

//==============================================================================
// Calculate TuneWord from output frequency
//==============================================================================
uint32 `$INSTANCE_NAME`_CalcStep( double freq )
{ 
    uint32 tmp = (uint32) (freq * Tdiv + 0.5);           // calculate tune word
    return tmp;
}


//==============================================================================
//     Return actual output frequency
//==============================================================================
double `$INSTANCE_NAME`_GetOutpFreq()  
{	  
    //double OutputFreq  = (((double) tune_word) / DDS_RESOLUTION ) * `$INSTANCE_NAME`_CLOCK_FREQ;  
    
    //double OutputFreq  = (((double) `$INSTANCE_NAME`_ReadStep()) / DDS_RESOLUTION ) * `$INSTANCE_NAME`_CLOCK_FREQ;   
    //return OutputFreq;
    
    // todo: use Tdiv
    //return ((((double) `$INSTANCE_NAME`_ReadStep()) / DDS_RESOLUTION ) * `$INSTANCE_NAME`_CLOCK_FREQ);   
    
    return ((((double) aDDS_inc) / DDS_RESOLUTION ) * `$INSTANCE_NAME`_PresetClockFreq);   
}

//==============================================================================
//     Return current set frequency 
//==============================================================================
double `$INSTANCE_NAME`_GetFrequency() { return aFrequency; }


//==============================================================================
//      
//==============================================================================
uint8 `$INSTANCE_NAME`_GetEnable()
{   
    //todo: must be initialized first
    return aEnabled;
}


/*
//==============================================================================
// Fill buffer with output data
//==============================================================================

void DDS_FillBuffer(uint16 start, uint16 length )
{     
    uint8  adr = 0;     // sine table address
    uint8  val = 0;     // sine value
    
    uint16 i;
    for (i=0; i<length; i++)
    {
        dds_acc += dds_inc;
        adr = (dds_acc >> 24);
        val = sine8[adr];

        WaveGen_1_Buffer[start+i] = val;
    }    
        
}

//==============================================================================
// Fill buffer with output data
//==============================================================================

void DDS_FillHalfBuffer()
{     
    uint8  adr = 0;     // 8-bit sine table address
    uint16 adr10 = 0;     // 16-bit sine table address
    uint8  val = 0;     // sine value
    
    // the speed of buffer update determins maximum allowed Clock_DDS
    // code must be optimized to fill 1/2 buffer before isr fires again
    // increase BUS_CLOCK and WaveGen_1_NumSamples if jitter observed
    
    uint16 i;
    for (i=0; i<WaveGen_1_NumRefHigh; i++) // todo: this assumes equal halves sizes
    {
        dds_acc += dds_inc;
        
        
        // 8-bit table->
        //adr = (dds_acc >> 24);      // 8-bit table
        ///val = sine8[adr];           // todo: set amplitude here
        ///val = 128 + (DDS_Amplitude * (sine8[adr]-128) / 256 ); // ampl=int8 [-128..127]; //ok
        //val = 128 + ((DDS_Amplitude * (sine8[adr]-128)) >> 8); // ampl=int8 [-128..127]; // ok
        
        
        // 10-bit table-> //todo: this code is slower because sine10 takes 52 ticks!
        adr10 = (dds_acc >> 22);      
        ///val = 128 + (DDS_Amplitude * mysine(adr10) / 2048 ); // ampl=uint8 [0..255]; //ok
        val = 128 + ((DDS_Amplitude * mysine(adr10)) >> 11 ); // ampl=uint8 [0..255]; //ok
        
        

        
        * WriteBufferIndex++  = val; // populate buffer and advance position
        ///if (WriteBufferIndex >= pBuffer + BUF_NumSamples) WriteBufferIndex = pBuffer;  // reset pointer
        if (WriteBufferIndex >= &WaveGen_1_Buffer[0] + WaveGen_1_NumSamples) WriteBufferIndex = &WaveGen_1_Buffer[0];  // reset pointer
    }    
        
}
*/


/*
#if ((`$INSTANCE_NAME`_DataFormat==_SimpleBinary) & (`$INSTANCE_NAME`_DataPacketSize==1) )
//==============================================================================
// Unique function to send data in Simple Binary format, int8 (uint8), N=1       
// this is simplest and fastest format, which does not need
// neither header nor checksum to transmit the data
//==============================================================================
void `$INSTANCE_NAME`_Plot(`=$Args`)
{
    // Format:      Simple Binary
    // Channels:    1
    // Number Type: `=$data_t`
    // Endianness:  Little Endian

    `=$UART`_PutChar(V1);    // unique 8-bit case
}
#endif
*/




/* [] END OF FILE */


/*
    #if   (`$INSTANCE_NAME`_DataType == _int8)
        //uint8 adr = (aDDS_acc >> 24);   // for table size 2^8
        ///`$data_t` val = sine8[adr] ^ 0x80;
        ///`$INSTANCE_NAME`_Reference = sine8[aDDS_acc >> 24] ^0x80;
        //int8  val = sine8[adr] ^0x80;
        //int8 val2 = sine8[(uint8)(adr + 64u)] ^0x80;
        //`$INSTANCE_NAME`_Reference = val2;

        
        //uint16 adr = (aDDS_acc >> 22);   // for table size 2^10=1024
        //int8 val = 128 + (aAmplitude * mysine(adr) / 1024 ); // ampl_v=int8 [-128..127];
        //int8 val = (aAmplitude * mysine(adr) >> 12 ); // ampl_v=int8 [-128..127];
        //val ^= 0x80; // ampl_v=int8 [-128..127];
        //uint8 val =  ((mysine(adr) ^ 0x400) >> 3; // 0-255;
        //uint8 val = ( ( aAmplitude * mysine(adr) ) ^ 0x400) >> 3; // [-128..127]
        //int8 val = (int32)(1024 + 4 * (mysine(adr))) >> 11; // [-128..127]
        //int8 val = (int32)(   0 + 5 * (mysine(adr))) >> 11; // [-128..127]
        
        //aAmplitude=3; //debug
        uint16 phase = (aDDS_acc >> 22);   // for table size 2^10=1024
        const uint16 half_pi = 256;
        int32 offset = (aAmplitude & 0x01)? 0 : 1024;
        int16 val1 = (int32) (offset + aAmplitude * mysine(phase)) >> 11; // [-128..127]
        int16 val2 = (int32) (offset + aAmplitude * mysine(phase+half_pi)) >> 11; // [-128..127]

        // table size 2048
        //uint16 phase = (aDDS_acc >> 21);   // for table size 2^11=2048
        //const uint16 half_pi = 512;
        //int32 offset = (aAmplitude & 0x01)? 0 : 1024;
        //int16 val1 = (int32) (offset + aAmplitude * mysine(phase)) >> 11; // [-128..127]
        //int16 val2 = (int32) (offset + aAmplitude * mysine(phase+half_pi)) >> 11; // [-128..127]


        
    #elif (`$INSTANCE_NAME`_DataType == _uint8)
        //uint8 adr = (aDDS_acc >> 24);
        //int8 val = sine8[adr];
        //int8 val2 = sine8[(uint8)(adr + 64u)];
        //`$INSTANCE_NAME`_Reference = val2;

        //uint16 adr = (aDDS_acc >> 22);   // for table size 2^10=1024
        //uint8 val = 128 + (aAmplitude * mysine(theta) / 1024 ); // ampl_v=int8 [-128..127];
        //uint8 val =  mysine(adr) >> 3; // 0-255;
        //uint8 val = (int32) (1024 + 4 * (mysine(adr) )) >> 11; // [0..255] // using unsigned mysine
        //uint8 val = (int32) (1024 + 2 * (mysine(adr)+1024 )) >> 11; // [0..255] // using signed mysine
        
        
        //aAmplitude=5; //debug
        uint16 adr = (aDDS_acc >> 22);   // for table size 2^10=1024
        uint16 val1 = (uint32) (1024 + aAmplitude * (mysine(adr)+1024)) >> 11; // [0..255] // using signed mysine
        uint16 val2 = (uint32) (1024 + aAmplitude * (mysine(adr+256)+1024)) >> 11; // cosin [0..255] // using signed mysine

        
        
    #elif (`$INSTANCE_NAME`_DataType == _int16)//todo: move to float?
        
        //aAmplitude=2048; //debug
        
        uint16  adr = (aDDS_acc >> 16);
        float32 Phase = (float32) (2 * M_PI * adr / 65536.0);
        `$data_t` val = 0.5*aAmplitude * sin(Phase);
        //`$data_t` val = round(aAmplitude * sin(Phase));
        
        int16 val2= 0.5*aAmplitude * cos(Phase);
        int16 val2 = round(0.5*aAmplitude * cos(Phase));
        `$INSTANCE_NAME`_Reference = val2;
      
    #elif (`$INSTANCE_NAME`_DataType == _uint16)
        
        //uint16  adr = (aDDS_acc >> 16);
        //float32 Phase = (float32) (2 * M_PI * adr / 65536.0);
        //`$data_t` val = aAmplitude + aAmplitude * sin(Phase);
        ///`$data_t` val = round(aAmplitude + aAmplitude * sin(Phase));
        
        
        aAmplitude=255; //debug
        uint16 adr = (aDDS_acc >> 22);   // for table size 2^10=1024
        uint16 val = (uint32) (1024 + aAmplitude * (mysine(adr)+1024)) >> 11; // [0..255] // using signed mysine
        uint16 val2= (uint32) (1024 + aAmplitude * (mysine(adr+256)+1024)) >> 11; // cosin [0..255] // using signed mysine
        `$INSTANCE_NAME`_Reference = val2;
        

    #endif
*/
